#pragma once

#include <max.h>
#include <iparamb2.h>
#include "ClassID.h"

namespace Ephere { namespace Plugins { namespace Autodesk { namespace Max { namespace Ornatrix
{

/** Defines various operations performed by the Edit Guides modifier.
	Use this class to modify guides in various ways programmatically.
*/
class EditGuidesModifierFPInterface: public ::FPMixinInterface
{
public:

	//! Puts Max into "Create Root" mode where user can plant new guides on top of the distribution object
	virtual void CreateRoot() = 0;

	//! Puts Max into "Import Spline" mode where user can pick a spline object to be added as a guide
	virtual void ImportSpline() = 0;

	//! Puts Max into "Control point change" mode where user can select where on a guide they want control points to be positioned
	virtual void CPChange() = 0;

	//! Cuts all selected strands at the current position of control points. The new tips will be at the control points.
	virtual void CutStrands() = 0;

	//! Creates a braid out of selected guides using the current braid configuration
	virtual void BraidStrands() = 0;

	//! Captures the current guide shape from the top of the stack into this modifier. Guides will be edited to look like current modifier stack result.
	virtual void AdoptFromStack() = 0;

	// Brushes:

	/** Sets the type of current brush.
		@param brush The brush type to set. 0 - Select, 1 - Move, 2 - Rotate, 3 - Scale, 4 - Cut, 5 - Grow/Shrink, 6 - Paint, 7 - Grab, 8 - Smooth, 9 - Revert
	*/
	virtual void SetBrushType( int brush ) = 0;

	//! Gets the type of current brush. 0 - Select, 1 - Move, 2 - Rotate, 3 - Scale, 4 - Cut, 5 - Grow/Shrink, 6 - Paint, 7 - Grab, 8 - Smooth, 9 - Revert
	virtual int GetBrushType() const = 0;

	//! Sets the width and height of the current brush. Height parameter is only relevant when shape is rectangular
	virtual void SetBrushSize( float width, float height ) = 0;

	//! Gets the width or radius of the current brush
	virtual float GetBrushWidth() const = 0;

	//! Gets the height of the current brush, when shape is rectangular
	virtual float GetBrushHeight() const = 0;

	//! Sets the shape of the current brush. 0 - Circle, 1 - Rectangle, 2 - Messy
	virtual void SetBrushShape( int shape ) = 0;

	//! Gets the shape of the current brush. 0 - Circle, 1 - Rectangle, 2 - Messy
	virtual int GetBrushShape() const = 0;

	//! Sets the affect strength of current brush 
	virtual void SetBrushStrength( float value ) = 0;

	//! Gets the affect strength of current brush 
	virtual float GetBrushStrength() const = 0;

	//! Gets whether only currently selected guides will be affected by brushing 
	virtual BOOL GetAffectSelectedOnly() const = 0;

	//! Sets whether only currently selected guides will be affected by brushing  
	virtual void SetAffectSelectedOnly( BOOL value ) = 0;

	//! Gets whether only guides whose roots are under the current brush will be affected
	virtual BOOL GetAffectByRootsOnly() const = 0;

	//! Sets whether only guides whose roots are under the current brush will be affected 
	virtual void SetAffectByRootsOnly( BOOL value ) = 0;

	//! Gets whether vertex weights will determine how much brushing affects each vertex 
	virtual BOOL GetRespectVertexWeights() const = 0;

	//! Sets whether vertex weights will determine how much brushing affects each vertex  
	virtual void SetRespectVertexWeights( BOOL value ) = 0;

	// Root operations:

	//! Gets the current selection level. 0 - Normal, 1 - Root, 2 - Strand, 3 - Brush
	virtual int GetSelectionLevel() const = 0;

	//! Sets the current selection level . 0 - Normal, 1 - Root, 2 - Strand, 3 - Brush
	virtual void SetSelectionLevel( int level ) = 0;

	//! Hides selected roots, and unhides any previously hidden ones
	virtual void HideSelectedRoots( BOOL unhidePrevious ) = 0;

	//! Hides specified roots  and unhides any previously hidden ones
	virtual void HideRoots( ::BitArray* sel, BOOL unhidePrevious ) = 0;

	//! Unhides all roots in this object 
	virtual void UnhideAllRoots() = 0;

	//! Gets an array specifying which roots are currently hidden
	virtual ::BitArray* GetHidden() const = 0;

	//! Freezes selected roots, and unfreezes any previously frozen ones  
	virtual void FreezeSelectedRoots( BOOL unhidePrevious ) = 0;

	//! Freezes specified roots, and unfreezes any previously frozen ones
	virtual void FreezeRoots( ::BitArray* sel, BOOL unfreezePrevious ) = 0;

	//! Unfreezes all roots
	virtual void UnfreezeAllRoots() = 0;

	//! Gets an array specifying frozen roots 
	virtual ::BitArray* GetFrozen() const = 0;

	//! Marks specified roots as selected 
	virtual void SelectRoots( ::BitArray* sel ) = 0;

	//! Gets an array specifying selected roots 
	virtual ::BitArray* GetSelected() const = 0;	

	//! Gets whether the guides gizmo is currently hidden or visible 
	virtual BOOL GetHideGuidesGizmo() const = 0;

	//! Sets whether the guides gizmo is currently hidden or visible
	virtual void SetHideGuidesGizmo( BOOL hide ) = 0;

	//! Gets whether backfacing roots will be selectable
	virtual BOOL GetCullBackfacingRoots() const = 0;

	//! Sets whether backfacing roots will be selectable
	virtual void SetCullBackfacingRoots( BOOL hide ) = 0;

	//! Gets whether backfacing roots are visible
	virtual BOOL GetHideBackfacingRoots() const = 0;

	//! Sets whether backfacing roots are visible
	virtual void SetHideBackfacingRoots( BOOL hide ) = 0;

	//! Gets whether each strand vertex is marked in viewport
	virtual BOOL GetShowVertexTicks() const = 0;

	//! Sets whether each strand vertex is marked in viewport
	virtual void SetShowVertexTicks( BOOL hide ) = 0;

	// Brush mode:
	virtual BOOL GetAdditiveMode() const = 0;
	virtual void SetAdditiveMode( BOOL ) = 0;
	virtual BOOL GetScaleMode() const = 0;
	virtual void SetScaleMode( BOOL ) = 0;

	// Channels:
	virtual int GetCurrentChannelIndex() const = 0;
	virtual BOOL GetCurrentChannelIndexIsRoot() const = 0;
	virtual void SetCurrentChannelIndex( int index, BOOL isRootChannel ) = 0;

	//! Pops up a dialog to prompt the user to create a new channel
	virtual void AddChannel() = 0;

	/** Creates a new guide channel
		@param rootChannel If true a root channel will be created. Otherwise a vertex channel will be created.
		@param name Name for the new channel
		@return The index of the newly created channel or -1 if there was an error
	*/
	virtual int CreateChannel( BOOL rootChannel, const MCHAR* name ) = 0;

	virtual void DeleteCurrentChannel( BOOL showConfirmDialog ) = 0;

	// Channel groups:
	virtual int GetChannelDisplayMode() const = 0;
	virtual void SetChannelDisplayMode( int ) = 0;

	/** Sets the value of specified channel for the selected entities
		@param isRootChannel Determines whether to set root or vertex channel value
		@param channelIndex Index of the channel for which to set value
		@param value The new value to assign
	*/
	virtual void SetChannelValueToSelection( BOOL isRootChannel, int channelIndex, float value ) = 0;

	/** Gets the shape of specified strand in object coordinates
	@param strandIndex Index of the strand
	@return Points, in object coordinates, defining the strand's shape
	*/
	virtual ::Tab<::Point3*>* GetStrandShape( int strandIndex ) = 0;

	/** Sets the shape of the specified strand to specified points in object coordinates
		@param strandIndex Index of the strand
		@param points A collection of points, in object coordinates, to assign to the strand
	*/
	virtual void SetStrandShape( int strandIndex, ::Tab<::Point3*>* points ) = 0;

	/** Gets the number of selection sets in this modifier */
	virtual int NamedSelectionSetCount() = 0;

	/** Adds a new named selection set based on current strand selection
		@param name Name to assign the new selection set
	*/
	virtual void AddNamedSelectionSetFromSelection( const MCHAR* name ) = 0;

	/** Deletes named selection set with specified name
		@param name Name of the selection set to delete
	*/
	virtual void DeleteNamedSelectionSet( const MCHAR* name ) = 0;

	/** Selects strands in specified named selection set
		@param name Name of the selection set
	*/
	virtual void ActivateNamedSelectionSet( const MCHAR* name ) = 0;

	/** Pushes out selected guides away from the distribution mesh
		@param selectedGuidesOnly When false all guides are pushed out, otherwise only selected ones
	*/
	virtual void PushAwayFromMesh( BOOL selectedGuidesOnly ) = 0;

	//! Use this method for retrieving this interface from an instance of a ground strands modifier
	static EditGuidesModifierFPInterface* GetInterface( ::InterfaceServer* obj )
	{
		return static_cast<EditGuidesModifierFPInterface*>( obj->GetInterface( ORNAEDITGUIDES_INTERFACE ) );
	}

protected:

	enum Functions
	{
		oxi_eg_create_root,
		oxi_eg_import_spline,
		oxi_eg_control_point,
		oxi_eg_cut_strands,
		oxi_eg_braid_strands,
		oxi_eg_adopt_from_stack,

		// Brushes:
		oxi_eg_set_brush_type,
		oxi_eg_get_brush_type,
		oxi_eg_set_brush_size,
		oxi_eg_get_brush_width,
		oxi_eg_get_brush_height,
		oxi_eg_set_brush_shape,
		oxi_eg_get_brush_shape,
		oxi_eg_set_brush_strength,
		oxi_eg_get_brush_strength,
		oxi_eg_get_affect_selected_only,
		oxi_eg_set_affect_selected_only,
		oxi_eg_get_affect_by_roots_only,
		oxi_eg_set_affect_by_roots_only,
		oxi_eg_get_respect_vertex_weights,
		oxi_eg_set_respect_vertex_weights,

		// Root operations:
		oxi_eg_get_selection_level,
		oxi_eg_set_selection_level,
		oxi_eg_select_roots,
		oxi_eg_get_selected,
		oxi_eg_get_hidden,
		oxi_eg_hide_selected,
		oxi_eg_hide_roots,
		oxi_eg_unhide_all,
		oxi_eg_get_frozen,
		oxi_eg_freeze_selected,
		oxi_eg_freeze_roots,
		oxi_eg_unfreeze_all,
		oxi_eg_get_hide_guides_gizmo,
		oxi_eg_set_hide_guides_gizmo,
		oxi_eg_get_cull_backfacing_roots,
		oxi_eg_set_cull_backfacing_roots,
		oxi_eg_get_hide_backfacing_roots,
		oxi_eg_set_hide_backfacing_roots,
		oxi_eg_get_show_vertex_ticks,
		oxi_eg_set_show_vertex_ticks,

		// Brush mode:
		oxi_eg_get_additive_mode,
		oxi_eg_set_additive_mode,
		oxi_eg_get_scale_mode,
		oxi_eg_set_scale_mode,

		// Channels:
		oxi_eg_get_current_channel_index,
		oxi_eg_get_current_channel_index_is_root,
		oxi_eg_set_current_channel_index,
		oxi_eg_add_channel,
		oxi_eg_create_channel,
		oxi_eg_delete_current_channel,

		// Channel groups:
		oxi_eg_get_channel_display_mode,
		oxi_eg_set_channel_display_mode,
		oxi_eg_set_channel_value_for_selection,

		Function_GetStrandShape,
		Function_SetStrandShape,

		Function_NamedSelectionSetCount,
		Function_AddNamedSelectionSetFromSelection,
		Function_DeleteNamedSelectionSet,
		Function_ActivateNamedSelectionSet,

		Function_PushAwayFromMesh
	};

private:

	BEGIN_FUNCTION_MAP
		// Root management:
		VFN_0( oxi_eg_create_root, CreateRoot );
		VFN_0( oxi_eg_import_spline, ImportSpline );
		VFN_0( oxi_eg_control_point, CPChange );
		VFN_0( oxi_eg_cut_strands, CutStrands );
		VFN_0( oxi_eg_braid_strands, BraidStrands );
		VFN_0( oxi_eg_adopt_from_stack, AdoptFromStack );

		// Brushes:
		VFN_1( oxi_eg_set_brush_type, SetBrushType, TYPE_INT );
		FN_0( oxi_eg_get_brush_type, TYPE_INT, GetBrushType );
		VFN_2( oxi_eg_set_brush_size, SetBrushSize, TYPE_FLOAT, TYPE_FLOAT );
		FN_0( oxi_eg_get_brush_width, TYPE_FLOAT, GetBrushWidth );
		FN_0( oxi_eg_get_brush_height, TYPE_FLOAT, GetBrushHeight );
		VFN_1( oxi_eg_set_brush_shape, SetBrushShape, TYPE_INT );
		FN_0( oxi_eg_get_brush_shape, TYPE_INT, GetBrushShape );
		VFN_1( oxi_eg_set_brush_strength, SetBrushStrength, TYPE_FLOAT );
		FN_0( oxi_eg_get_brush_strength, TYPE_FLOAT, GetBrushStrength );
		FN_0( oxi_eg_get_affect_selected_only, TYPE_BOOL, GetAffectSelectedOnly );
		VFN_1( oxi_eg_set_affect_selected_only, SetAffectSelectedOnly, TYPE_BOOL );
		FN_0( oxi_eg_get_affect_by_roots_only, TYPE_BOOL, GetAffectByRootsOnly );
		VFN_1( oxi_eg_set_affect_by_roots_only, SetAffectByRootsOnly, TYPE_BOOL );
		FN_0( oxi_eg_get_respect_vertex_weights, TYPE_BOOL, GetRespectVertexWeights );
		VFN_1( oxi_eg_set_respect_vertex_weights, SetRespectVertexWeights, TYPE_BOOL );

		// Root operations:
		VFN_1( oxi_eg_set_selection_level, SetSelectionLevel, TYPE_INT );
		FN_0( oxi_eg_get_selection_level, TYPE_INT, GetSelectionLevel );
		VFN_1( oxi_eg_select_roots, SelectRoots, TYPE_BITARRAY );
		FN_0( oxi_eg_get_selected, TYPE_BITARRAY, GetSelected );

		// |- Hide:
		FN_0( oxi_eg_get_hidden, TYPE_BITARRAY, GetHidden );
		VFN_1( oxi_eg_hide_selected, HideSelectedRoots, TYPE_BOOL );
		VFN_2( oxi_eg_hide_roots, HideRoots, TYPE_BITARRAY, TYPE_BOOL );
		VFN_0( oxi_eg_unhide_all, UnhideAllRoots );

		// |- Freeze:
		FN_0( oxi_eg_get_frozen, TYPE_BITARRAY, GetFrozen );
		VFN_1( oxi_eg_freeze_selected, FreezeSelectedRoots, TYPE_BOOL );
		VFN_2( oxi_eg_freeze_roots, FreezeRoots, TYPE_BITARRAY, TYPE_BOOL );
		VFN_0( oxi_eg_unfreeze_all, UnfreezeAllRoots );

		// |- Other properties:
		FN_0( oxi_eg_get_hide_guides_gizmo, TYPE_BOOL, GetHideGuidesGizmo );
		VFN_1( oxi_eg_set_hide_guides_gizmo, SetHideGuidesGizmo, TYPE_BOOL );
		FN_0( oxi_eg_get_cull_backfacing_roots, TYPE_BOOL, GetCullBackfacingRoots );
		VFN_1( oxi_eg_set_cull_backfacing_roots, SetCullBackfacingRoots, TYPE_BOOL );
		FN_0( oxi_eg_get_hide_backfacing_roots, TYPE_BOOL, GetHideBackfacingRoots );
		VFN_1( oxi_eg_set_hide_backfacing_roots, SetHideBackfacingRoots, TYPE_BOOL );
		FN_0( oxi_eg_get_show_vertex_ticks, TYPE_BOOL, GetShowVertexTicks );
		VFN_1( oxi_eg_set_show_vertex_ticks, SetShowVertexTicks, TYPE_BOOL );

		// Root add/remove:

		// Brush mode:
		FN_0( oxi_eg_get_additive_mode, TYPE_BOOL, GetAdditiveMode );
		VFN_1( oxi_eg_set_additive_mode, SetAdditiveMode, TYPE_BOOL );
		FN_0( oxi_eg_get_scale_mode, TYPE_BOOL, GetScaleMode );
		VFN_1( oxi_eg_set_scale_mode, SetScaleMode, TYPE_BOOL );

		// Channels:
		FN_0( oxi_eg_get_current_channel_index, TYPE_INT, GetCurrentChannelIndex );
		FN_0( oxi_eg_get_current_channel_index_is_root, TYPE_BOOL, GetCurrentChannelIndexIsRoot );
		VFN_2( oxi_eg_set_current_channel_index, SetCurrentChannelIndex, TYPE_INT, TYPE_BOOL );
		VFN_0( oxi_eg_add_channel, AddChannel );
		FN_2( oxi_eg_create_channel, TYPE_INT, CreateChannel, TYPE_BOOL, TYPE_STRING );
		VFN_1( oxi_eg_delete_current_channel, DeleteCurrentChannel, TYPE_BOOL );

		// Channel groups:
		FN_0( oxi_eg_get_channel_display_mode, TYPE_INT, GetChannelDisplayMode );
		VFN_1( oxi_eg_set_channel_display_mode, SetChannelDisplayMode, TYPE_INT );
		VFN_3( oxi_eg_set_channel_value_for_selection, SetChannelValueToSelection, TYPE_BOOL, TYPE_INT, TYPE_FLOAT );

		FN_1( Function_GetStrandShape, TYPE_POINT3_TAB, GetStrandShape, TYPE_INT );
		VFN_2( Function_SetStrandShape, SetStrandShape, TYPE_INT, TYPE_POINT3_TAB );
		FN_0( Function_NamedSelectionSetCount, TYPE_INT, NamedSelectionSetCount );
		VFN_1( Function_AddNamedSelectionSetFromSelection, AddNamedSelectionSetFromSelection, TYPE_STRING );
		VFN_1( Function_DeleteNamedSelectionSet, DeleteNamedSelectionSet, TYPE_STRING );
		VFN_1( Function_ActivateNamedSelectionSet, ActivateNamedSelectionSet, TYPE_STRING );

		VFN_1( Function_PushAwayFromMesh, PushAwayFromMesh, TYPE_BOOL );
	END_FUNCTION_MAP

};

} } } } } // Ephere::Plugins::Autodesk::Max::Ornatrix