#pragma once

#include "ClassID.h"
#include <max.h>

#define GetOrnaGuidesInterface(obj) ((GuidesFPInterface*)obj->GetInterface(ORNA_GUIDES_INTERFACE))

namespace Ephere { namespace Plugins { namespace Autodesk { namespace Max { namespace Ornatrix
{

class HairFPInterface;

/** Provides MaxScript access to guides.
	This class is typically retrieved via GuidesObjectFPInterface::GetGuides
*/
class GuidesFPInterface : public ::FPMixinInterface
{
public:

	::FPInterfaceDesc* GetDesc() override
	{
		return GetDescByID( ORNA_GUIDES_INTERFACE );
	}

	// Own:

	/** Gets the hair interface which can be used to manipulate the guides topology, geometry, and channel data among other things.
		@return Hair interface
	*/
	virtual HairFPInterface* AsHair() = 0;

	/** Sets the total number of strands thus resizing the guides object
		@param count Root count
		@param keep If true any previous roots will be kept if possible
	*/
	virtual void SetNumRoots( int count, BOOL keep ) = 0;

	/** Deletes specified strands from the guides object
		@param set Roots to delete
	*/
	virtual void DeleteRootSet( ::BitArray* set ) = 0;

	/** Sets the total number of vertices in guides
		@param count Vertex count
		@param keep If true previous vertices will be kept when possible
	*/
	virtual void SetNumVerts( int count, BOOL keep ) = 0;

	/** Appends specified vertices to the vertex array
		@param vertices Vertices to append
		@param count Vertex count
	*/
	virtual void AppendVerts( ::Tab<Point3*>* vertices, int count ) = 0;

	/** Deletes specified vertices
		@param set Vertices to delete
	*/
	virtual void DeleteVertSet( ::BitArray* set ) = 0;

	/** Sets the number of per-vertex channels
		@param count Per-vertex channel count
		@bool keep If true, previous channel data will be kept when possible
	*/
	virtual void VertexDataCount( int count, BOOL keep ) = 0;

	/** Gets single vertex data for specified channel and vertex.
		@param channelIndex Channel index
		@param vertexIndex Vertex index
		@return Channel value
	*/
	virtual float GetVertData( int channelIndex, int vertexIndex ) = 0;

	/** Sets single vertex data for specified channel and vertex
		@param channelIndex Channel index
		@param vertexIndex Vertex index
		@param value Channel value
	*/
	virtual void SetVertData( int channelIndex, int vertexIndex, float value ) = 0;

	/** Gets a single vertex data for specified channel, root, and point
		@param channelIndex Channel index
		@param rootIndex Strand index
		@param pointIndex Index of point on strand
	*/
	virtual float GetPointData( int channelIndex, int rootIndex, int pointIndex ) = 0;

	/** Sets a single vertex data for specified channel, root, and point
		@param channelIndex Channel index
		@param rootIndex Strand index
		@param pointIndex Index of point on strand
		@param value Channel value
	*/
	virtual void SetPointData( int channelIndex, int rootIndex, int pointIndex, float value ) = 0;

	/** Sets the vertex channel name for specified channel. This name appears in the UI.
		@param channelIndex Channel index
		@param name Name to assign
	*/
	virtual void SetVertChanName( int channelIndex, TSTR name ) = 0;

	/** Gets the vertex channel name for specified channel. This name appears in the UI.
		@param channelIndex Channel index
		@return Channel name
	*/
	virtual TSTR GetVertChanName( int channelIndex ) = 0;

	/** Delete a set of channels based on a bit array
		@param set Channels to delete
	*/
	virtual void DeleteVertChanSet( ::BitArray* set ) = 0;

	/** Sets the number of per-root channels
		@param count Per-root channel count
		@bool keep If true, previous channel data will be kept when possible
	*/
	virtual void RootDataCount( int count, BOOL keep ) = 0;

	/** Gets single root data for specified channel and root.
		@param channelIndex Channel index
		@param rootIndex Root index
		@return Channel value
	*/
	virtual float GetRootData( int channelIndex, int rootIndex ) = 0;

	/** Sets single root data for specified channel and root
		@param channelIndex Channel index
		@param rootIndex Root index
		@param value Channel value
	*/
	virtual void SetRootData( int channelIndex, int rootIndex, float value ) = 0;

	/** Sets the root channel name for specified channel. This name appears in the UI.
		@param channelIndex Channel index
		@param name Name to assign
	*/
	virtual void SetRootChanName( int channelIndex, TSTR name ) = 0;

	/** Gets the root channel name for specified channel. This name appears in the UI.
		@param channelIndex Channel index
		@return Channel name
	*/
	virtual TSTR GetRootChanName( int channelIndex ) = 0;

	/** Delete a set of channels based on a bit array
		@param set Channels to delete
	*/
	virtual void DeleteRootChanSet( ::BitArray* set ) = 0;

	/** Gets root selection value
		@param rootIndex Root index
		@return Selection value
	*/
	virtual float GetRootSel( int rootIndex ) = 0;

	/** Sets root selection value
		@param rootIndex Root index
		@param selection Selection value
	*/
	virtual void SetRootSel( int rootIndex, float selection ) = 0;

	/** Gets a vertex weight value. This may be used when editing or brushing guides to affect points differently.
		@param vertexIndex Vertex index
		@return Vertex weight
	*/
	virtual float GetWeight( int vertexIndex ) = 0;

	/** Sets a vertex weight value. This may be used when editing or brushing guides to affect points differently.
		@param vertexIndex Vertex index
		@param weight Vertex weight
	*/
	virtual void SetWeight( int vertexIndex, float weight ) = 0;

	/** Sets a display flag
		@param flag Display flag
	*/
	virtual void SetDispFlag( int flag ) = 0;

	/** Gets a display flag
		@param flag Display flag
		@return Flag value
	*/
	virtual int	GetDispFlag( int flag ) = 0;

	/** Clears a display flag
		@param flag Display flag
	*/
	virtual void ClearDispFlag( int flag ) = 0;

	/** Gets the current binary selection state of strands
		@return Strand selection
	*/
	virtual ::BitArray* GetStrandSelection() = 0;

	/** Sets the current binary selection state of strands
		@param selection Strand selection
	*/
	virtual void SetStrandSelection( ::BitArray* selection ) = 0;

protected:

	enum Members
	{
		ogi_as_hair,
		ogi_set_num_roots, ogi_delete_root_set,
		ogi_set_num_verts, ogi_append_verts, ogi_delete_vert_set,
		ogi_set_num_vert_data, ogi_get_vert_data, ogi_set_vert_data, ogi_get_point_data, ogi_set_point_data,
		ogi_set_vert_chan_name, ogi_get_vert_chan_name, ogi_delete_vert_chan_set,
		ogi_set_num_root_data, ogi_get_root_data, ogi_set_root_data, ogi_set_root_chan_name,
		ogi_get_root_chan_name, ogi_delete_root_chan_set,
		ogi_get_root_sel, ogi_set_root_sel, ogi_get_weight, ogi_set_weight,
		ogi_set_disp_flag, ogi_get_disp_flag, ogi_clear_disp_flag,
		Member_GetStrandSelection,
		Member_SetStrandSelection,

		MemberCount
	};

private:

	BEGIN_FUNCTION_MAP
		FN_0( ogi_as_hair, TYPE_INTERFACE, AsHair );
		VFN_2( ogi_set_num_roots, SetNumRoots, TYPE_INT, TYPE_BOOL );
		VFN_1( ogi_delete_root_set, DeleteRootSet, TYPE_BITARRAY );
		VFN_2( ogi_set_num_verts, SetNumVerts, TYPE_INT, TYPE_BOOL );
		VFN_2( ogi_append_verts, AppendVerts, TYPE_POINT3_TAB, TYPE_INT );
		VFN_1( ogi_delete_vert_set, DeleteVertSet, TYPE_BITARRAY );
		VFN_2( ogi_set_num_vert_data, VertexDataCount, TYPE_INT, TYPE_BOOL );
		FN_2( ogi_get_vert_data, TYPE_FLOAT, GetVertData, TYPE_INT, TYPE_INT );
		VFN_3( ogi_set_vert_data, SetVertData, TYPE_INT, TYPE_INT, TYPE_FLOAT );
		FN_3( ogi_get_point_data, TYPE_FLOAT, GetPointData, TYPE_INT, TYPE_INT, TYPE_INT );
		VFN_4( ogi_set_point_data, SetPointData, TYPE_INT, TYPE_INT, TYPE_INT, TYPE_FLOAT );
		VFN_2( ogi_set_vert_chan_name, SetVertChanName, TYPE_INT, TYPE_TSTR );
		FN_1( ogi_get_vert_chan_name, TYPE_TSTR, GetVertChanName, TYPE_INT );
		VFN_1( ogi_delete_vert_chan_set, DeleteVertChanSet, TYPE_BITARRAY );
		VFN_2( ogi_set_num_root_data, RootDataCount, TYPE_INT, TYPE_BOOL );
		FN_2( ogi_get_root_data, TYPE_FLOAT, GetRootData, TYPE_INT, TYPE_INT );
		VFN_3( ogi_set_root_data, SetRootData, TYPE_INT, TYPE_INT, TYPE_FLOAT );
		VFN_2( ogi_set_root_chan_name, SetRootChanName, TYPE_INT, TYPE_TSTR );
		FN_1( ogi_get_root_chan_name, TYPE_TSTR, GetRootChanName, TYPE_INT );
		VFN_1( ogi_delete_root_chan_set, DeleteRootChanSet, TYPE_BITARRAY );
		FN_1( ogi_get_root_sel, TYPE_FLOAT, GetRootSel, TYPE_INT );
		VFN_2( ogi_set_root_sel, SetRootSel, TYPE_INT, TYPE_FLOAT );
		FN_1( ogi_get_weight, TYPE_FLOAT, GetWeight, TYPE_INT );
		VFN_2( ogi_set_weight, SetWeight, TYPE_INT, TYPE_FLOAT );
		VFN_1( ogi_set_disp_flag, SetDispFlag, TYPE_INT );
		FN_1( ogi_get_disp_flag, TYPE_INT, GetDispFlag, TYPE_INT );
		VFN_1( ogi_clear_disp_flag, ClearDispFlag, TYPE_INT );
		FN_0( Member_GetStrandSelection, TYPE_BITARRAY, GetStrandSelection );
		VFN_1( Member_SetStrandSelection, SetStrandSelection, TYPE_BITARRAY );
	END_FUNCTION_MAP

};

} } } } } // Ephere::Plugins::Autodesk::Max::Ornatrix