#pragma once

#include "ClassID.h"
#include <max.h>

namespace Ephere { namespace Plugins { namespace Autodesk { namespace Max { namespace Ornatrix
{

class GuidesFPInterface;

/** Provides base level access to hair and guides which flows through Ornatrix pipeline
*/
class HairFPInterface: public ::FPMixinInterface
{
public:

	::FPInterfaceDesc* GetDesc() override
	{
		return GetDescByID( ORNA_HAIR_INTERFACE );
	}

	/** Gets total number of roots
		@return Total number of roots
	*/
	virtual int NumRoots() = 0;

	/** Set number of roots and all associated data (specified in flags)
		@param count Number of roots
		@param keep If true previous roots will not be cleared
	*/
	virtual void SetNumRoots( int count, BOOL keep ) = 0;

	/** Gets whether this hair use root topology data
		@return true if this hair uses root topology data
	*/
	virtual BOOL UsesStrandTopology() = 0;

	/** Get starting index into vertex array of specified root
		@return Starting index into vertex array of specified root
	*/
	virtual int GetRootGID( int r ) = 0;

	/** Get number of vertices associated with specified root
		@return Number of vertices associated with specified root
	*/
	virtual int GetRootGNum( int r ) = 0;

	/** Set starting index into vertex array of specified root
		@param rootIndex Root index
		@param vertexIndex Vertex index
	*/
	virtual void SetRootGID( int rootIndex, int vertexIndex ) = 0;

	/** Set number of vertices associated with specified root
		@param rootIndex Root index
		@param pointCount Point count
	*/
	virtual void SetRootGNum( int rootIndex, int pointCount ) = 0;

	/** Gets whether this hair use root transform data
		@return true if this hair uses root transform data
	*/
	virtual BOOL HasRootTM() = 0;

	/** Gets strand matrix associated with root
		@param rootIndex Root index
		@return Matrix associated with root
	*/
	virtual ::Matrix3* GetRootTM( int rootIndex ) = 0;

	/** Set a new matrix associated with root
		@param rootIndex Root index
		@param transform Matrix associated with root
	*/
	virtual void SetRootTM( int rootIndex, ::Matrix3 transform ) = 0;

	/** Gets whether this hair use surface dependance data
		@return true if this hair uses surface dependance data
	*/
	virtual BOOL HasSurfDep() = 0;

	/** Gets the distribution mesh face index on which specified root is planted
		@param rootIndex Root index
		@return Face index
	*/
	virtual int GetSurfDepFace( int rootIndex ) = 0;

	/** Gets the barycentric coordinate on the distribution mesh face on which specified root is planted
		@param rootIndex Root index
		@return Barycentric coordinate
	*/
	virtual ::Point3* GetSurfDepBary( int rootIndex ) = 0;

	/** Sets the distribution mesh face index on which specified root is to be located
		@param rootIndex Root index
		@param faceIndex Face index
	*/
	virtual void SetSurfDepFace( int rootIndex, int faceIndex ) = 0;

	/** Sets the barycentric coordinate on the distribution mesh face on which specified root is to be located
		@param rootIndex Root index
		@param barycentricCoordinate Barycentric coordinate
	*/
	virtual void SetSurfDepBary( int rootIndex, ::Point3 barycentricCoordinate ) = 0;

	/** Gets whether this hair use guide dependance data
		@return true if this hair uses guide dependance data
	*/
	virtual BOOL HasGuideDep() = 0;

	/** Gets the index of the n'th closest guide to specified root, if current hair is generated from guides
		@param rootIndex Root index
		@param closestGuideIndex The index of the closest guide
		@return Guide index
	*/
	virtual int GetGuideDepCSR( int rootIndex, int closestGuideIndex ) = 0;

	/** Gets the distance from the n'th closest guide to specified root, if current hair is generated from guides
		@param rootIndex Root index
		@param closestGuideIndex The index of the closest guide
		@return Distance to guide
	*/
	virtual float GetGuideDepDSR( int rootIndex, int closestGuideIndex ) = 0;

	/** Sets the index of the n'th closest guide to specified root, if current hair is generated from guides
		@param rootIndex Root index
		@param closestGuideIndex The index of the closest guide
		@param guideIndex Guide index
	*/
	virtual void SetGuideDepCSR( int rootIndex, int closestGuideIndex, int guideIndex ) = 0;

	/** Sets the distance from the n'th closest guide to specified root, if current hair is generated from guides
		@param rootIndex Root index
		@param closestGuideIndex The index of the closest guide
		@param distance Distance to guide
	*/
	virtual void SetGuideDepDSR( int rootIndex, int closestGuideIndex, float distance ) = 0;

	/** Computes root transform matrix for specified strand
		@param rootIndex Root index
		@return Strand to object transform
	*/
	virtual ::Matrix3* ComputeRootTM( int rootIndex ) = 0;

	/** Updates root transforms using surface dependency and distribution mesh normals
		@param startingRootIndex Index of the first root
		@param count Number of roots for which to compute transforms
	*/
	virtual void UpdateRootTM( int startingRootIndex, int count ) = 0;

	/** Gets vertex index for the first point in specified root
		@param rootIndex Root index
		@return Vertex index
	*/
	virtual int VertId( int rootIndex ) = 0;

	/** Gets the number of points in specified strand
		@param rootIndex Root index
		@return Point count
	*/
	virtual int NumPts( int rootIndex ) = 0;

	/** Gets the global number of points per strand
		@return Point count per strand
	*/
	virtual int GetGlobSegNum() = 0;

	/** Sets the global number of points per strand
		@param count Point count per strand
	*/
	virtual void SetGlobSegNum( int count ) = 0;

	/** Gets the total number of hair vertices
		@return Vertex count
	*/
	virtual int NumVerts() = 0;

	/** Sets the number of vertices and resizes all connected data
		@param count Vertex count
		@param keep If true previous vertices will be kept as much as possible
	*/
	virtual void SetNumVerts( int count, BOOL keep ) = 0;

	/** Append a vertex array to the end of vertex list and update lookup tables
		@param vertices Vertex positions
		@param count Number of vertices
	*/
	virtual void AppendVerts( ::Tab<Point3*>* vertices, int count ) = 0;

	/** Gets vertex in local strand space
		@param vertexIndex Vertex index
		@return Vertex position
	*/
	virtual ::Point3* GetVert( int vertexIndex ) = 0;

	/** Gets vertex in object space
		@param vertexIndex Vertex index
		@return Vertex position
	*/
	virtual ::Point3* GetVertT( int vertexIndex ) = 0;

	/** Sets vertex in local strand space
		@param vertexIndex Vertex index
		@param position Vertex position
	*/
	virtual void SetVert( int vertexIndex, ::Point3 position ) = 0;

	/** Sets vertex in object space
		@param vertexIndex Vertex index
		@param position Vertex position
	*/
	virtual void SetVertT( int vertexIndex, ::Point3 position ) = 0;

	/** Gets strand point in local strand space
		@param rootIndex Strand index
		@param pointIndex Point index on strand
		@return Point position
	*/
	virtual ::Point3* GetPoint( int rootIndex, int pointIndex ) = 0;

	/** Gets strand point in object space
		@param rootIndex Strand index
		@param pointIndex Point index on strand
		@return Point position
	*/
	virtual ::Point3* GetPointT( int rootIndex, int pointIndex ) = 0;

	/** Sets strand point in local strand space
		@param rootIndex Strand index
		@param pointIndex Point index on strand
		@param position Point position
	*/
	virtual void SetPoint( int rootIndex, int pointIndex, ::Point3 position ) = 0;

	/** Sets strand point in object space
		@param rootIndex Strand index
		@param pointIndex Point index on strand
		@param position Point position
	*/
	virtual void SetPointT( int rootIndex, int pointIndex, ::Point3 position ) = 0;

	/** Gets whether class uses vertex lookup tables
		@return true if class uses vertex lookup tables
	*/
	virtual int HasVertInd() = 0;

	/** Gets root index associated with specified vertex index
		@param vertexIndex Vertex index
		@return Root index
	*/
	virtual int GetVertInd( int vertexIndex ) = 0;

	/** Sets root index associated with specified vertex index
		@param vertexIndex Vertex index
		@param rootIndex Root index
	*/
	virtual void SetVertInd( int vertexIndex, int rootIndex ) = 0;

	/** Updates vertex lookup tables based on available data
	*/
	virtual void RemapVertInd() = 0;

	/** Gets the texture coordinate of specified point on specified strand
		@param strandIndex Index of the strand
		@param pointIndex Index of a vertex on the strand
		@return Texture coordinate at specified position and strand
	*/
	virtual ::Point3* GetPointUVW( int strandIndex, int pointIndex ) = 0;

	/** Gets the texture coordinate at the root of the specified strand
		@param strandIndex Strand index
		@return Texture coordinate
	*/
	virtual ::Point3* GetVertUVW( int strandIndex ) = 0;

	/** Validates the bounding box of this object
	*/
	virtual void UpdateBoundingBox() = 0;

	/** Gets the minimum coordinate of the hair's bounding box
		@return Minimum corner coordinate
	*/
	virtual ::Point3* GetBoundingBoxMin() = 0;

	/** Gets the maximum coordinate of the hair's bounding box
		@return Maximum corner coordinate
	*/
	virtual ::Point3* GetBoundingBoxMax() = 0;

	/** Saves this hair into an .oxh file
		@param filePath File path
	*/
	virtual void Save( TSTR filePath ) = 0;

	/** Loads this hair from an .oxh file
		@param filePath File path
	*/
	virtual void Load( TSTR filePath ) = 0;

	/** If this hair structure indirectly depends on guides, this will return an interface to those guides. You can see which guide each hair depends on by reading the guide dependence data.
		@return Guides interface
	*/
	virtual GuidesFPInterface* GetGuides( ) = 0;

	/** If you would like your hair automatically deform with a mesh, you can assign a mesh here and then call UpdateRootTM() function to have all root transforms 'glued' to the surface.
		You also need to aset up surface dependence data for this to work. Mesh has to be valid when roots are updates, so take good care not to garbage-collect or delete the provided object by accident.
		@param obj Distribution mesh object
	*/
	virtual void SetDistrObj( ::ReferenceTarget* obj ) = 0;

	//! Use this method for retrieving this interface from an instance of a hair interface
	static HairFPInterface* GetInterface( ::InterfaceServer* obj )
	{
		return static_cast<HairFPInterface*>( obj->GetInterface( ORNA_HAIR_INTERFACE ) );
	}

protected:

	enum Members
	{
		ohi_num_roots,
		ohi_set_num_roots,
		ohi_has_roots,
		ohi_get_root_gid,
		ohi_get_root_gnum,
		ohi_set_root_gid,
		ohi_set_root_gnum,
		ohi_has_root_tm,
		ohi_get_root_tm,
		ohi_set_root_tm,
		ohi_has_surf_dep,
		ohi_get_surf_dep_face,
		ohi_get_surf_dep_bary,
		ohi_set_surf_dep_face,
		ohi_set_surf_dep_bary,
		ohi_has_guide_dep,
		ohi_get_guide_dep_csr,
		ohi_get_guide_dep_dsr,
		ohi_set_guide_dep_csr,
		ohi_set_guide_dep_dsr,
		ohi_compute_root_tm,
		ohi_update_root_tm,
		ohi_vert_id, ohi_num_pts,
		ohi_get_glob_seg_num,
		ohi_set_glob_seg_num,
		ohi_num_verts,
		ohi_set_num_verts,
		ohi_append_verts,
		ohi_get_vert,
		ohi_get_vert_t,
		ohi_set_vert,
		ohi_set_vert_t,
		ohi_get_point,
		ohi_get_point_t,
		ohi_set_point,
		ohi_set_point_t,
		ohi_has_vert_ind,
		ohi_get_vert_ind,
		ohi_set_vert_ind,
		ohi_remap_vert_ind,
		Member_GetPointUVW,
		ohi_get_vert_uvw,
		ohi_update_bounding_box,
		ohi_get_bounding_box_min,
		ohi_get_bounding_box_max,
		ohi_save, ohi_load,
		ohi_get_guides,
		ohi_set_distr_obj,

		MemberCount
	};

private:

	BEGIN_FUNCTION_MAP
		FN_0( ohi_num_roots, TYPE_INT, NumRoots );
		VFN_2( ohi_set_num_roots, SetNumRoots, TYPE_INT, TYPE_BOOL );
		FN_0( ohi_has_roots, TYPE_BOOL, UsesStrandTopology );
		FN_1( ohi_get_root_gid, TYPE_INT, GetRootGID, TYPE_INT );
		FN_1( ohi_get_root_gnum, TYPE_INT, GetRootGNum, TYPE_INT );
		VFN_2( ohi_set_root_gid, SetRootGID, TYPE_INT, TYPE_INT );
		VFN_2( ohi_set_root_gnum, SetRootGNum, TYPE_INT, TYPE_INT );
		FN_0( ohi_has_root_tm, TYPE_BOOL, HasRootTM );
		FN_1( ohi_get_root_tm, TYPE_MATRIX3, GetRootTM, TYPE_INT );
		VFN_2( ohi_set_root_tm, SetRootTM, TYPE_INT, TYPE_MATRIX3 );
		FN_0( ohi_has_surf_dep, TYPE_BOOL, HasSurfDep );
		FN_1( ohi_get_surf_dep_face, TYPE_INT, GetSurfDepFace, TYPE_INT );
		FN_1( ohi_get_surf_dep_bary, TYPE_POINT3, GetSurfDepBary, TYPE_INT );
		VFN_2( ohi_set_surf_dep_face, SetSurfDepFace, TYPE_INT, TYPE_INT );
		VFN_2( ohi_set_surf_dep_bary, SetSurfDepBary, TYPE_INT, TYPE_POINT3 );
		FN_0( ohi_has_guide_dep, TYPE_BOOL, HasGuideDep );
		FN_2( ohi_get_guide_dep_csr, TYPE_INT, GetGuideDepCSR, TYPE_INT, TYPE_INT );
		FN_2( ohi_get_guide_dep_dsr, TYPE_FLOAT, GetGuideDepDSR, TYPE_INT, TYPE_INT );
		VFN_3( ohi_set_guide_dep_csr, SetGuideDepCSR, TYPE_INT, TYPE_INT, TYPE_INT );
		VFN_3( ohi_set_guide_dep_dsr, SetGuideDepDSR, TYPE_INT, TYPE_INT, TYPE_FLOAT );
		FN_1( ohi_compute_root_tm, TYPE_MATRIX3, ComputeRootTM, TYPE_INT );
		VFN_2( ohi_update_root_tm, UpdateRootTM, TYPE_INT, TYPE_INT );
		FN_1( ohi_vert_id, TYPE_INT, VertId, TYPE_INT );
		FN_1( ohi_num_pts, TYPE_INT, NumPts, TYPE_INT );
		FN_0( ohi_get_glob_seg_num, TYPE_INT, GetGlobSegNum );
		VFN_1( ohi_set_glob_seg_num, SetGlobSegNum, TYPE_INT );
		FN_0( ohi_num_verts, TYPE_INT, NumVerts );
		VFN_2( ohi_set_num_verts, SetNumVerts, TYPE_INT, TYPE_BOOL );
		VFN_2( ohi_append_verts, AppendVerts, TYPE_POINT3_TAB, TYPE_INT );
		FN_1( ohi_get_vert, TYPE_POINT3, GetVert, TYPE_INT );
		FN_1( ohi_get_vert_t, TYPE_POINT3, GetVertT, TYPE_INT );
		VFN_2( ohi_set_vert, SetVert, TYPE_INT, TYPE_POINT3 );
		VFN_2( ohi_set_vert_t, SetVertT, TYPE_INT, TYPE_POINT3 );
		FN_2( ohi_get_point, TYPE_POINT3, GetPoint, TYPE_INT, TYPE_INT );
		FN_2( ohi_get_point_t, TYPE_POINT3, GetPointT, TYPE_INT, TYPE_INT );
		VFN_3( ohi_set_point, SetPoint, TYPE_INT, TYPE_INT, TYPE_POINT3 );
		VFN_3( ohi_set_point_t, SetPointT, TYPE_INT, TYPE_INT, TYPE_POINT3 );
		FN_0( ohi_has_vert_ind, TYPE_BOOL, HasVertInd );
		FN_1( ohi_get_vert_ind, TYPE_INT, GetVertInd, TYPE_INT );
		VFN_2( ohi_set_vert_ind, SetVertInd, TYPE_INT, TYPE_INT );
		VFN_0( ohi_remap_vert_ind, RemapVertInd );
		FN_2( Member_GetPointUVW, TYPE_POINT3, GetPointUVW, TYPE_INT, TYPE_INT );
		FN_1( ohi_get_vert_uvw, TYPE_POINT3, GetVertUVW, TYPE_INT );
		VFN_0( ohi_update_bounding_box, UpdateBoundingBox );
		FN_0( ohi_get_bounding_box_min, TYPE_POINT3, GetBoundingBoxMin );
		FN_0( ohi_get_bounding_box_max, TYPE_POINT3, GetBoundingBoxMax );
		VFN_1( ohi_save, Save, TYPE_TSTR );
		VFN_1( ohi_load, Load, TYPE_TSTR );
		FN_0( ohi_get_guides, TYPE_INTERFACE, GetGuides );
		VFN_1( ohi_set_distr_obj, SetDistrObj, TYPE_REFTARG );
	END_FUNCTION_MAP

};

} } } } } // Ephere::Plugins::Autodesk::Max::Ornatrix