#pragma once

class BitArray;

#define I_USERINTERFACE 0x0000ffff

namespace Ephere { namespace Plugins { namespace Autodesk { namespace Max { namespace Ornatrix
{

enum
{
	//! Interface ID to use with a 3dsmax Animatable::GetInterface( ULONG ) method to get a pointer to IHair
	IGuidesInterfaceID = I_USERINTERFACE + 51035
};

// forward declares:
class IHair;

/**
 *  Ornatrix guides interface class. Used in conjunction with IHair class to add selection, channel, and weight functionality to hair objects.
 */
class IGuides
{
public:

	// Vertex data:

	//! Gets the number of vertex data channels
	virtual unsigned VertexDataCount() const = 0;

	/** Sets number of vertex data channels. This will also resize the vertex data array.
		@param numChans New number of vertex channels
		@param keep If TRUE vertex channel data is not reset
	*/
	virtual void VertexDataCount( unsigned numChans, bool keep = true ) = 0;

	/** Retreives single vertex data for specified channel and vertex.
		@param channel Channel index
		@param vertexIndex Vertex index
		@return Associated vertex data
	*/
	virtual float GetVertData( unsigned channel, unsigned vertexIndex ) const = 0;

	/** Retreives vertex data pointer for specified channel and root.
		@see GetVertData()
		@param rootIndex Root index
		@return Associated vertex data pointer
	*/
	virtual float* GetVertData( unsigned rootIndex ) = 0;

	/** Set the value of vertex data for specified channel and vertex.
		@see GetVertData()
		@param channel Channel index
		@param vertexIndex Vertex index
		@param value New value for data
	*/
	virtual void SetVertData( unsigned channel, unsigned vertexIndex, float value ) = 0;

	/** Retreives single vertex data for specified channel, root, and vertex.
		@param chan Channel index
		@param r Root index
		@param v Vertex index relative to the root
		@return Associated vertex data
	*/
	virtual float GetVertData( unsigned chan, unsigned r, unsigned v ) const = 0;

	/** Set the value of vertex data for specified channel, root, and vertex.
		@param chan Channel index
		@param r Root index
		@param v Vertex index relative to the root
		@param data New value for data
	*/
	virtual void SetVertData( unsigned chan, unsigned r, unsigned v, float value ) = 0;

	/** Set the vertex channel name for specified channel. This name appears in the UI.
		@param chan Channel index
		@param name Null-terminated name string, maximum of 16 characters
	*/
	virtual void SetVertChanName( unsigned chan, const wchar_t* name ) = 0;

	/** Get the vertex channel name for specified channel. This name appears in the UI.
		@param chan Channel index
		@return Null-terminated name string, maximum of 16 characters
	*/
	virtual const wchar_t* GetVertChanName( unsigned chan ) const = 0;

	/** Delete a set of channels based on a bit array.
		@param set Set which is used for deletion (1-delete channel, 0-leave channel)
	*/
	virtual void DeleteVertChanSet( const ::BitArray& set ) = 0;

	// Root data:

	virtual void RootDataCount( unsigned numChans, bool keep = true ) = 0;

	//! Gets number of root data channels
	virtual unsigned RootDataCount() const = 0;

	virtual float GetRootData( unsigned channelIndex, unsigned rootIndex ) const = 0;

	virtual void SetRootData( unsigned channelIndex, unsigned rootIndex, float value ) = 0;

	virtual void DeleteRootChanSet( const ::BitArray& set ) = 0;

	virtual void SetRootChanName( unsigned channelIndex, const wchar_t* name ) = 0;

	virtual const wchar_t* GetRootChanName( unsigned channelIndex ) const = 0;

	// Selection:

	/** Gets the scalar selection value of a strand.
		Binary root selection set via GetRootSelection() must be set for this value to have any effect.
		@param rootIndex Index of the strand
		@return Scalar selection value
	*/
	virtual float GetRootSel( unsigned rootIndex ) const = 0;

	/** Sets the scalar selection value of a strand.
		Binary root selection set via GetRootSelection() must be set for this value to have any effect.
		@param rootIndex Index of the strand
		@param value The new selection value
	*/
	virtual void SetRootSel( unsigned rootIndex, float value ) = 0;

	virtual float GetWeight( unsigned vertexIndex ) const = 0;

	virtual void SetWeight( unsigned vertexIndex, float weight ) = 0;

	//! Gets the hair interface referenced by these guides
	virtual const IHair& AsHair() const = 0;

	inline IHair& AsHair()
	{
		return const_cast<IHair&>( static_cast<const IGuides&>( *this ).AsHair() );
	}

	template<class T>
	inline T& AsHair()
	{
		return static_cast<T&>( AsHair() );
	}

	/** Gets the binary selection states of individual strands.
		A strand is only considered selected if its binary selection is true, even if its scalar selection set via SetRootSel(...) method is non-zero.
	*/
	virtual const ::BitArray& GetRootSelection() const = 0;

	/** Gets the binary selection states of individual strands.
		A strand is only considered selected if its binary selection is true, even if its scalar selection set via SetRootSel(...) method is non-zero.
	*/
	inline ::BitArray& GetRootSelection()
	{
		return const_cast<::BitArray&>( static_cast<const IGuides&>( *this ).GetRootSelection() );
	}

};

} } } } } // Ephere::Plugins::Autodesk::Max::Ornatrix