#pragma once

// forward declarations:
class Point3;
class Matrix3;
class Box3;
class GraphicsWindow;
class Material;
class HitRegion;
class Mesh;

// forward declarations:
template<class T> class Tab;

#define I_USERINTERFACE 0x0000ffff

namespace Ephere { namespace Plugins { namespace Autodesk { namespace Max { namespace Ornatrix
{

enum
{
	//! Interface ID to use with a 3dsmax Animatable::GetInterface( ULONG ) method to get a pointer to IHair
	IHairInterfaceID = I_USERINTERFACE + 4414
};

// forward declarations:
struct StrandTopology;
struct SurfaceDependency;
struct GuideDependency;
class IGuides;

/**
 *  Ornatrix hair interface class. Base 'shell' for all ornatrix strand objects that contain hair data.
 */
class IHair
{
public:

	/** Copy hair from another class
		@see DeepCopy()
	*/
	virtual IHair& operator=( const IHair& fromHair ) = 0;

	/** Appends geometry frames from a given hair structure. Note that hair topology in
		both structures has to be identical for this to work.
		@see InsertFrames()
		@param fromHair Source hair class
		@return Modifier hair
	*/
	virtual IHair& operator+=( const IHair& fromHair ) = 0;

	/** Copy a shape of the array of strand vertices from this class into a given array of
		strand vertices. Number of points in each array is independent.
		@see VertID()
		@see NumPts()
		@see CopyHairShape()
		@param vertexIndex Index of first vertex into verts array
		@param pointCount Number of points in source vertex array
		@param destination Destination array
		@param destinationCount Number of points in destination vertex array
		@return true if successful
	*/
	virtual bool CopyStrandShape( unsigned vertexIndex, unsigned pointCount, ::Point3* destination, unsigned destinationCount, const ::Matrix3& destinationTransform, float frame = 0.0f ) = 0;

	/** Modify the hair shape of given hair class to match the shape of this classes hair.
		This function does not care about the topology of destination hair. If number of
		roots exceeds or preceeds the required number only the available data is copied.
		Strand transformation of both structures is taken into account.
		@param destination Destination hair structure
		@return true if successful
	*/
	virtual bool CopyHairShape( IHair& destination, float frame = 0.0f ) = 0;

	///////////////////////////////////////////////////////////////////////////////////////////
	// Memory
	///////////////////////////////////////////////////////////////////////////////////////////

	//! Deletes all data used by this class
	virtual void FreeAll() = 0;

	////! Deletes all root data
	//virtual void FreeRoots() = 0;

	////! Deletes all vertex data
	//virtual void FreeVerts() = 0;

	////! Allocates all the data needed to store roots (based on flags)
	//virtual void AllocRoots() = 0;

	////! Allocates all the data needed to store vertices (based on flags)
	//virtual void AllocVerts() = 0;

	///////////////////////////////////////////////////////////////////////////////////////////
	// Roots and topology
	///////////////////////////////////////////////////////////////////////////////////////////

	/**
	* Returns number of roots (strands) in this class.
	*/
	virtual unsigned NumRoots() const = 0;

	/** Set number of roots and all associated data (specified in flags)
	 *	@param ct New number of roots
	 *	@param keep If TRUE previous data is kept, otherwise it is deleted
	 */
	virtual void SetNumRoots( unsigned ct, bool keep = false ) = 0;

	/**
	* @return true if each strands in this structure can have a strands to have varying vertex counts
	*/
	virtual bool UsesStrandTopology() const = 0;

	/**
	* Determines whether to use same vertex count for all strands (if false) or have ability for strands to have varying vertex count (if true)
	* @see UsesStrandTopology()
	* @param on true to have ability for strands to have varying vertex count
	*/
	virtual void UsesStrandTopology( bool on ) = 0;

	/**
	* @return true if a special table is maintained that allows quickly determining a strand's index based on specified index inside of it
	*/
	virtual bool UsesVertexLookupTable() const = 0;

	/**
	* Determines whether to use a special table is maintained that allows quickly determining a strand's index based on specified index inside of it.
	* This requires some extra memory but can greatly speed up certain algorithms requiring this functionality.
	* @see UsesVertexLookupTable()
	* @param on true to allow lookup table
	*/
	virtual void UsesVertexLookupTable( bool on ) = 0;

	virtual bool UsesInstancedStrands() const = 0;

	virtual void UsesInstancedStrands( bool on ) = 0;

	virtual bool UsesPerRootTransformations() const = 0;

	virtual void UsesPerRootTransformations( bool on ) = 0;

	virtual bool KeepsSurfaceDependency() const = 0;

	virtual void KeepsSurfaceDependency( bool on ) = 0;

	virtual bool KeepsGuideDependency() const = 0;

	virtual void KeepsGuideDependency( bool on ) = 0;

	virtual bool UsesPolarInterpolation() const = 0;

	virtual void UsesPolarInterpolation( bool on ) = 0;

	virtual bool UsesTopologyOrientation() const = 0;

	virtual void UsesTopologyOrientation( bool on ) = 0;

	virtual bool UsesGlobalPerStrandVertexCount() const = 0;

	virtual void UsesGlobalPerStrandVertexCount( bool on ) = 0;

	virtual bool HiddenStrands() const = 0;

	virtual void HiddenStrands( bool on ) = 0;

	virtual bool HideBackfacingStrands() const = 0;

	virtual void HideBackfacingStrands( bool on ) = 0;

	/**
	* Copies all flags from specified source hair class into this one
	* @param source Hair structure from which to copy flags
	*/
	inline void CopyFlags( const IHair& source )
	{
		UsesStrandTopology( source.UsesStrandTopology() );
		UsesVertexLookupTable( source.UsesVertexLookupTable() );
		UsesInstancedStrands( source.UsesInstancedStrands() );
		UsesPerRootTransformations( source.UsesPerRootTransformations() );
		KeepsSurfaceDependency( source.KeepsSurfaceDependency() );
		KeepsGuideDependency( source.KeepsGuideDependency() );
		UsesPolarInterpolation( source.UsesPolarInterpolation() );
		UsesTopologyOrientation( source.UsesTopologyOrientation() );
		UsesGlobalPerStrandVertexCount( source.UsesGlobalPerStrandVertexCount() );
		HiddenStrands( source.HiddenStrands() );
		HideBackfacingStrands( source.HideBackfacingStrands() );
	}

	//! Sets all flags in this hair structure to false
	inline void ClearFlags()
	{
		UsesStrandTopology( false );
		UsesVertexLookupTable( false );
		UsesInstancedStrands( false );
		UsesPerRootTransformations( false );
		KeepsSurfaceDependency( false );
		KeepsGuideDependency( false );
		UsesPolarInterpolation( false );
		UsesTopologyOrientation( false );
		UsesGlobalPerStrandVertexCount( false );
		HiddenStrands( false );
		HideBackfacingStrands( false );
	}

	/**
	* @see UsesStrandTopology()
	* @param i Index of root
	* @return Root topology structure
	*/
	virtual const StrandTopology& GetRoot( unsigned index ) const = 0;

	/**
	* @see UsesStrandTopology()
	* @param i Index of root
	* @return Root topology structure
	*/
	inline StrandTopology& GetRoot( unsigned index )
	{
		return const_cast<StrandTopology&>( static_cast<const IHair&>( *this ).GetRoot( index ) );
	}

	/**
	* @return Non-zero if class uses per-strand root transforms
	*/
	virtual bool HasRootTM() const = 0;

	/**
	 *	@param i Root index
	 *	@return Matrix associated with root
	 */
	virtual ::Matrix3 GetRootTM( unsigned rootIndex ) const = 0;

	/**
	 *	@param i Root index
	 *	@param tm Reference to matrix storage
	 *	@return true if succesfull
	 */
	virtual bool GetRootTM( unsigned rootIndex, ::Matrix3& tm ) const = 0;

	/**
	* @see HasRootTM()
	* @param rootIndex Index of root
	* @param tm Matrix to set
	*/
	virtual void SetRootTM( unsigned rootIndex, const ::Matrix3& tm ) = 0;

	/**
	* @return Non-zero if class has per-strand surface dependancy
	*/
	virtual bool HasSurfDep() const = 0;

	/**
	* @see HasSurfDep()
	* @param rootIndex Index of root
	* @return Surface dependancy for specified root
	*/
	virtual const SurfaceDependency& GetSurfDep( unsigned rootIndex ) const = 0;

	/**
	* @see HasSurfDep()
	* @param rootIndex Index of root
	* @return Surface dependancy for specified root
	*/
	inline SurfaceDependency& GetSurfDep( unsigned rootIndex )
	{
		return const_cast<SurfaceDependency&>( static_cast<const IHair&>( *this ).GetSurfDep( rootIndex ) );
	}

	/**
	* @return Non-zero if class has per-strand guide dependancy
	*/
	virtual bool HasGuideDep() const = 0;

	/**
	* @see HasGuideDep()
	* @param i Index of root
	* @return Guide dependancy for specified root
	*/
	virtual const GuideDependency& GetGuideDep( unsigned rootIndex ) const = 0;

	/**
	* @see HasGuideDep()
	* @param i Index of root
	* @return Guide dependancy for specified root
	*/
	inline GuideDependency& GetGuideDep( unsigned rootIndex )
	{
		return const_cast<GuideDependency&>( static_cast<const IHair&>( *this ).GetGuideDep( rootIndex ) );
	}

	/** Copy root data from specified arrays, if NULL array is not copied.
	 *	@param from Root topology array
	 *	@param fromTM Root transform array
	 *	@param fromSD Surface dependency array
	 *	@param fromGD Guide dependency array
	 */
	virtual void CopyRootsFrom( const ::Tab<StrandTopology>* from, const ::Tab<Matrix3>* fromTM = 0, const ::Tab<SurfaceDependency>* fromSD = 0, const ::Tab<GuideDependency>* fromGD = 0 ) = 0;

	/** Computes root transform matrix for specified strand
	 *	@param rootIndex Root index
	 *	@param result Reference to matrix storage
	 */
	virtual void ComputeRootTM( unsigned rootIndex, ::Matrix3& result ) const = 0;

	/** Updates root transforms using surface dependency and distribution mesh normals
	 *	@param rfrom Start root index
	 *	@param rnum Number of roots to update
	 */
	virtual void UpdateRootTM( unsigned rfrom = 0, int rnum = -1 ) = 0;

	/**
	* @param r Index of root
	* @return Index into vertex array for the first point in specified root
	*/
	virtual unsigned VertId( unsigned rootIndex ) const = 0;

	/**
	* @param rootIndex Index of root
	* @return Number of points in specified root
	*/
	virtual unsigned NumPts( unsigned rootIndex ) const = 0;

	/**
	* @param num Number of points per strand to set
	*/
	virtual void SetGlobSegNum( unsigned num ) = 0;

	/**
	* @return Global number of points per strand (if OHF_GLOB_SEG_NUM is set)
	*/
	virtual unsigned GetGlobSegNum() const = 0;

	// Vertices and geometry
	/**
	* @return Total number of hair vertices
	* @param absolute Determines whether to return all vertices stored in the hair structure, or number of vertices stored just for one frame. Typically (without animation storage) this parameter will not affect the result.
	*/
	virtual unsigned NumVerts( bool absolute = false ) const = 0;

	/** Sets the number of vertices and resizes all connected data (based on flags)
	 *	@param vnum New number of vertices
	 *	@param keep If TRUE previous vertex information will be kept.
	 */
	virtual void SetNumVerts( unsigned vnum, bool keep = false ) = 0;

	/** Append a vertex array to the end of vertex list and update lookup tables (if used)
	 *	@param p Pointer to the vertex array to append
	 *	@param num Number of element in vertex array
	 */
	virtual void AppendVerts( const Point3* vertices, unsigned vertexCount ) = 0;

	/**
	* @param i Index of vertex
	* @return Position of specified vertex
	*/
	virtual const ::Point3& GetVert( unsigned vertexIndex ) const = 0;

	/**
	* @param i Index of vertex
	* @return Position of specified vertex
	*/
	inline ::Point3& GetVert( unsigned vertexIndex )
	{
		return const_cast<::Point3&>( static_cast<const IHair&>( *this ).GetVert( vertexIndex ) );
	}

	virtual ::Point3 GetVert( unsigned i, float frame ) const = 0;

	/**
	* @param rootIndex Index of root
	* @return First vertex for specified root
	*/
	virtual const ::Point3* GetVerts( unsigned rootIndex ) const = 0;

	/**
	* @param rootIndex Index of root
	* @return First vertex for specified root
	*/
	inline ::Point3* GetVerts( unsigned rootIndex )
	{
		return const_cast<::Point3*>( static_cast<const IHair&>( *this ).GetVerts( rootIndex ) );
	}

	/**
	* @param i Index of vertex
	* @param v Position to set
	*/
	virtual void SetVert( unsigned i, const ::Point3& v ) = 0;

	/**
	* @param rootIndex Index of root
	* @param vertexIndex Index of point on strand
	* @return Vertex position
	*/
	virtual const ::Point3& GetPoint( unsigned rootIndex, unsigned vertexIndex ) const = 0;

	/**
	* @param rootIndex Index of root
	* @param vertexIndex Index of point on strand
	* @return Vertex position
	*/
	inline ::Point3& GetPoint( unsigned rootIndex, unsigned vertexIndex )
	{
		return const_cast<::Point3&>( static_cast<const IHair&>( *this ).GetPoint( rootIndex, vertexIndex ) );
	}

	//! Gets a point in object coordinates
	inline ::Point3 GetPointInObjectCoordinates( unsigned rootIndex, unsigned pointIndex ) const
	{
		::Point3 result = GetPoint( rootIndex, pointIndex );
		if( UsesPerRootTransformations() )
		{
			result = result * GetRootTM( rootIndex );
		}
		
		return result;
	}

	virtual ::Point3 GetPoint( unsigned r, unsigned i, float frame ) const = 0;

	/**
	* @return Non-zero if class uses vertex lookup tables
	*/
	virtual bool HasVertInd() const = 0;

	/**
	* @param v Index of vertex
	* @param ind Lookup value to set
	*/
	virtual void SetVertInd( unsigned v, unsigned ind ) = 0;

	/**
	* @param v Index of vertex
	* @return Root index for specified vertex
	*/
	virtual unsigned GetVertInd( unsigned vertexIndex ) const = 0;

	/**
	* @param i Index of vertex
	* @return Root where specified vertex exists
	*/
	virtual StrandTopology* RootFromVert( unsigned i ) = 0;

	/** Updates vertex lookup tables based on available data
	*/
	virtual void RemapVertInd() = 0;

	///////////////////////////////////////////////////////////////////////////////////////////
	// UVW Texture Mapping
	///////////////////////////////////////////////////////////////////////////////////////////

	//! Gets the number of mapping channels present in this hair structure
	virtual unsigned MappingChannelCount() const = 0;

	/** Computes per-strand UV coordinates based on available data. This class is fast because it
	 *	computes all UVW data at once, otherwise GetUVW() function can be used.
	 *	@param result Array where UVW coords should be stored
	 *	@param mappingChannelIndex Index of mapping channel
	 *	@return true if succesful
	 */
	virtual bool ComputeUV( ::Tab<UVVert>& result, unsigned mappingChannelIndex = 0 ) const = 0;

	/** Computes per-strand UV coordinates for a single root
	 *	@param rootIndex Root index
	 *	@param vertexIndex Optional vertex index. If nonzero W coordinate is set to i/numPts
	 *	@param mappingChannelIndex Index of mapping channel
	 *	@return Computed UVW coordinate
	 */
	virtual ::UVVert GetUVW( unsigned rootIndex, unsigned pointIndex = 0, unsigned mappingChannelIndex = 0 ) const = 0;

	/**
	 *	@param vertexIndex Vertex index
	 *	@param mappingChannelIndex Index of mapping channel
	 *	@return Computed UVW coordinate
	 */
	virtual ::UVVert GetUVW( unsigned vertexIndex, unsigned mappingChannelIndex = 0 ) const = 0;

	///////////////////////////////////////////////////////////////////////////////////////////
	// Frames and animation
	///////////////////////////////////////////////////////////////////////////////////////////

	/**
	@return Total number of geometry frames in this hair structure
	*/
	virtual unsigned NumFrames() const = 0;

	/** Sets the number of frames in this hair structure 
		@param keep if true current frame data is kept
	*/
	virtual void SetNumFrames( unsigned numFrames, bool keep = false ) = 0;

	virtual void InsertFrames( const IHair& fromHair ) = 0;

	virtual void DeleteFrames( unsigned from, int range = -1 ) = 0;

	// Drawing and picking
	/** Draw individual strands into provided viewport.
		@param r Strand root index
		@param gw Viewport pointer
		@param hit If TRUE this function is used to hit-test the strand
		@param vtm If not NULL strand is drawn shaded with normals perpendicular to screen
		@return If hit is nonzero returns true if hit, and false otherwise
	*/
	virtual bool DrawStrand( unsigned r, ::GraphicsWindow* gw, bool hit, ::Matrix3* vtm = 0 ) = 0;

	/** Pick function to test if mouse cursor is over the hair.
		@return True if mouse is over the hair.
	*/
	virtual bool select( ::GraphicsWindow* gw, ::Material* ma, ::HitRegion* hr, bool abortOnHit = false, int numMat = 1 ) = 0;

	/** Interactive display function that will draw this hair into a viewport.

	This function uses display flags to test for rendering properties. It also uses sub-object flags to
	determine whether things like roots should be drawn.
	*/
	virtual void render( ::GraphicsWindow* gw, ::Material* ma, ::RECT* rp, int compFlags, int numMat = 1, ::Matrix3* vtm = 0 ) = 0;

	// Channels
	/** Copies just the pointers to the data specified by channels from another class.
		@param ahair Source hair class
		@param channels Which channel data to copy
	*/
	virtual void ShallowCopy( const IHair& source, ULONG_PTR channels ) = 0;

	/** Copies all class data using specified channel from another class
		@param ahair Source hair class
		@param channels Which channel data to copy
	*/
	virtual void DeepCopy( const IHair& source, ULONG_PTR channels ) = 0;

	/** Creates a new copy of all class data without erasing old data
		@param channels Channels which should be allocated and copied
	*/
	virtual void NewAndCopyChannels( ULONG_PTR channels ) = 0;

	/** Free data from specified channels from this class
	*/
	virtual void FreeChannels( ULONG_PTR channels, int zeroOthers = 1 ) = 0;

	// Bounding box:

	//! Updates bounding box for this hair class.
	virtual void UpdateBoundingBox() = 0;

	/**	Accessor to the hair bounding box
		@param box Box structure to fill with current bounding box
	*/
	virtual void GetBoundingBox( ::Box3& box ) const = 0;

	// Distribution mesh and source hair:

	//! Gets distribution mesh to which this hair applies or 0 if none is specified
	virtual ::Mesh* DistributionMesh() const = 0;

	//! Sets distribution mesh to which this hair applies or 0 if none is specified
	virtual void DistributionMesh( ::Mesh* value ) = 0;

	//! Gets guides which control this hair or 0 if none are assigned
	virtual IHair* Guides() const = 0;

	//! Sets guides which control this hair or 0 if none are assigned
	virtual void Guides( IHair* value ) = 0;


	// Guide conversion:

	//! If this hair represents guides an IGuides interface is returned, otherwise 0
	virtual const IGuides* AsGuides() const = 0;

	//! If this hair represents guides an IGuides interface is returned, otherwise 0
	inline IGuides* AsGuides()
	{
		return const_cast<IGuides*>( static_cast<const IHair&>( *this ).AsGuides() );
	}

	/** Gets guides for this hair at specified nesting level (guides can have their own guides)
		@param level Nesting level of the guides to get. For example, level 2 is analogous to this->Guides()->Guides()
	*/
	inline const IHair* GetGuidesByNestingLevel( unsigned level = 0 ) const
	{
		const IHair* result = this;
		for( unsigned index = 0; index < level; ++index )
		{
			if( result->Guides() != 0 )
			{
				result = result->Guides();
			}
			else
			{
				// Specified guide level is too deep
				return result;
			}
		}

		return result;
	}

	/** Gets guides for this hair at specified nesting level (guides can have their own guides)
		@param level Nesting level of the guides to get. For example, level 2 is analogous to this->Guides()->Guides()
	*/
	inline IHair* GetGuidesByNestingLevel( unsigned level = 0 )
	{
		return const_cast<IHair*>( static_cast<const IHair&>( *this ).GetGuidesByNestingLevel( level ) );
	}

	// Channel values:

	virtual float GetChannelVal( unsigned type, unsigned ind, unsigned root, unsigned point ) const = 0;

};

class IHair2: public IHair
{
public:

	/** Allows temporarily overriding the hair's surface dependency
		@param surfaceDependencies Pointer to an array of surface dependency values set to override internal values. 0 will reset the override to use internal values again.
	*/
	virtual void SetSurfaceDependencyOverride( const SurfaceDependency* surfaceDependencies = 0 ) = 0;

	// Repeated here from IHair to avoid hiding and warnings
	virtual void CopyRootsFrom( const ::Tab<StrandTopology>* from, const ::Tab<Matrix3>* fromTM = 0, const ::Tab<SurfaceDependency>* fromSD = 0, const ::Tab<GuideDependency>* fromGD = 0 ) = 0;

	virtual void CopyRootsFrom( const IHair& other, bool copyTopology = true, bool copyTransforms = true, bool copySurfaceDependencies = true, bool copyGuideDependencies = true ) = 0;

	/** Copy root data from specified arrays, if 0 array is not copied.
	 * @param rootCount Number of roots in each provided array
	 * @param sourceTopology Root topology array
	 * @param sourceTransforms Root transform array
	 * @param sourceSurfaceDependencies Surface dependency array
	 * @param sourceGuideDependencies Guide dependency array
	 */
	virtual void CopyRootsFrom( unsigned rootCount, const StrandTopology* sourceTopology, const ::Matrix3* sourceTransforms = 0, const SurfaceDependency* sourceSurfaceDependencies = 0, const GuideDependency* sourceGuideDependencies = 0 ) = 0;

	/** Computes per-strand UV coordinates based on available data. This class is fast because it
	*	computes all UVW data at once, otherwise GetUVW() function can be used.
	*	@param destination Array of UV coordinates where UVW coords should be stored. This must have the same size as the strand count in this hair structure.
	*	@param mappingChannelIndex Index of mapping channel
	*	@return true if succesful
	*/
	virtual bool GetStrandUVCoordinates( ::UVVert* destination, unsigned mappingChannelIndex = 0 ) const = 0;

};

class IHair3 : public IHair2
{
public:

	virtual inline void CopyFlags( const IHair3& source )
	{
		IHair2::CopyFlags( source );
		UsesCustomTextureCoordinates( source.UsesCustomTextureCoordinates() );
	}

	virtual inline void ClearFlags()
	{
		IHair2::ClearFlags();
		UsesCustomTextureCoordinates( false );
	}

	virtual ::Texmap* GetDisplacementMap() const = 0;

	virtual void SetDisplacementMap( ::Texmap* map ) = 0;

	virtual float GetDisplacementOffset() const = 0;

	virtual void SetDisplacementOffset( float offsetValue ) = 0;

	virtual unsigned GetDisplacementMapChannel() const = 0;

	virtual void SetDisplacementMapChannel( unsigned mappingChannelIndex = 0 ) = 0;

	/** Gets whether this hair object uses custom UVW coordinates set via SetTextureCoordinate function. If false, texture coordinates will be auto-generated from
	per-root based on surface dependency
	@return true if using custom texture coordinates
	*/
	virtual bool UsesCustomTextureCoordinates() const = 0;

	/** Sets whether this hair object uses custom UVW coordinates set via SetTextureCoordinate function. If false, texture coordinates will be auto-generated from
	per-root based on surface dependency
	@param value true to use custom texture coordinates
	*/
	virtual void UsesCustomTextureCoordinates( bool value ) = 0;

	/** Sets custom texture coordinate. This is useful only when UsesCustomTextureCoordinates() is true.
	@param strandIndex Index of the strand for which to set the coordinate
	@param pointIndex Index of the point on strand
	@param coordinate Texture coordinate to set
	*/
	virtual void SetTextureCoordinate( unsigned strandIndex, unsigned pointIndex, const ::Point3& coordinate ) = 0;

	virtual int GetMaterialIndex( unsigned strandIndex ) const = 0;

	virtual void SetMaterialIndex( unsigned strandIndex, int value ) const = 0;

};

} } } } } // Ephere::Plugins::Autodesk::Max::Ornatrix